package edu.sc.seis.cormorant.seismogram;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Collections;
import java.util.GregorianCalendar;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TimeZone;

import org.omg.CORBA.NO_IMPLEMENT;

import edu.iris.Fissures.FissuresException;
import edu.iris.Fissures.Time;
import edu.iris.Fissures.IfNetwork.ChannelId;
import edu.iris.Fissures.IfSeismogramDC.DataCenterCallBack;
import edu.iris.Fissures.IfSeismogramDC.DataCenterPOA;
import edu.iris.Fissures.IfSeismogramDC.LocalSeismogram;
import edu.iris.Fissures.IfSeismogramDC.RequestFilter;
import edu.iris.Fissures.model.MicroSecondDate;
import edu.iris.Fissures.model.TimeInterval;
import edu.iris.Fissures.model.UnitImpl;
import edu.iris.Fissures.network.ChannelIdUtil;
import edu.iris.Fissures.seismogramDC.LocalSeismogramImpl;
import edu.iris.Fissures.seismogramDC.RequestFilterUtil;
import edu.sc.seis.fissuresUtil.chooser.ThreadSafeSimpleDateFormat;
import edu.sc.seis.fissuresUtil.exceptionHandler.GlobalExceptionHandler;
import edu.sc.seis.fissuresUtil.mseed.FissuresConvert;
import edu.sc.seis.fissuresUtil.time.ReduceTool;
import edu.sc.seis.seisFile.mseed.Btime;
import edu.sc.seis.seisFile.mseed.BtimeRange;
import edu.sc.seis.seisFile.mseed.DataRecord;
import edu.sc.seis.seisFile.mseed.SeedFormatException;
import edu.sc.seis.seisFile.mseed.SeedRecord;
import edu.sc.seis.seisFile.mseed.Utility;

public class BudLightDataCenter extends DataCenterPOA {

    public BudLightDataCenter(File dataDir) {
        this.dataDir = dataDir;
    }

    private File dataDir;

    public RequestFilter[] available_data(RequestFilter[] request) {
        List<RequestFilter> out = new ArrayList<RequestFilter>();
        for (int r = 0; r < request.length; r++) {
            List<String> filenameList = formatFiles(request[r]);
            MicroSecondDate beginDate = null;
            for (String f : filenameList) {
                File file = new File(dataDir, f);
                logger.debug("available check "+file.getPath());
                if (file.exists() && beginDate == null) {
                    out.add(request[r]);
                    break;
                }
            }
        }
        return out.toArray(new RequestFilter[0]);
    }

    public String request_seismograms(RequestFilter[] a_filterseq,
                                      DataCenterCallBack a_client,
                                      boolean long_lived,
                                      Time expiration_time) throws FissuresException {
        throw new NO_IMPLEMENT();
    }


    public LocalSeismogram[] retrieve_seismograms(RequestFilter[] request) throws FissuresException {
        List<LocalSeismogramImpl> out = new ArrayList<LocalSeismogramImpl>();
        Map<String, List<RequestFilter>> splitRF = RequestFilterUtil.splitByChannel(Arrays.asList(request));
        for (String key : splitRF.keySet()) {
            out.addAll(processOneChannel(splitRF.get(key)));
        }
        LocalSeismogramImpl[] seis =  out.toArray(new LocalSeismogramImpl[0]);
        seis = ReduceTool.merge(seis);
        return seis;
    }
    
    /** Processes a list of requests, all of which must be from the same channel.
     * 
     */
    public List<LocalSeismogramImpl> processOneChannel(List<RequestFilter> request) throws FissuresException {
        List<LocalSeismogramImpl> out = new ArrayList<LocalSeismogramImpl>();
        List<DataRecord> allFromFile = retrieveDataRecords(request);
        List<List<DataRecord>> contiguous = Utility.breakContiguous(allFromFile);
        for (List<DataRecord> list : contiguous) {
            try {
                LocalSeismogramImpl seis = FissuresConvert.toFissures(list);
                seis.channel_id = request.get(0).channel_id; // respect the input request channel id
                out.add(seis);
            } catch(SeedFormatException e) {
                GlobalExceptionHandler.handle("Bad data for request: "+request.size()+" "+RequestFilterUtil.toString(request), e);
            }
        }
        logger.debug("try local for "+ChannelIdUtil.toStringNoDates(request.get(0).channel_id)+" "+out.size());
        return out;
    }

    public List<DataRecord> retrieveDataRecords(List<RequestFilter> request) {
        return retrieveDataRecords(request, "");
    }
    
    public List<DataRecord> retrieveDataRecords(List<RequestFilter> request, String fileSuffix) {
        List<DataRecord> allFromFile = new ArrayList<DataRecord>();
        Set<String> filenameSet = new HashSet<String>();
        List<BtimeRange> ranges = new ArrayList<BtimeRange>();
        for (RequestFilter rf : request) {
            filenameSet.addAll(formatFiles(rf));
            ranges.add(new BtimeRange(FissuresConvert.getBtime(new MicroSecondDate(rf.start_time)),
                                      FissuresConvert.getBtime(new MicroSecondDate(rf.end_time))));
        }
        List<String> sortedFilenames = new ArrayList<String>();
        sortedFilenames.addAll(filenameSet);
        Collections.sort(sortedFilenames);
        for (String f : sortedFilenames) {
            List<DataRecord> fromOneFile = retrieveDataRecords(f+fileSuffix, ranges);
            allFromFile.addAll(fromOneFile);
        }
        return allFromFile;
    }

    public List<DataRecord> retrieveDataRecords(String filename, List<BtimeRange> ranges) {
        List<DataRecord> allFromFile = new ArrayList<DataRecord>();
        File file = new File(dataDir, filename);
        if (file.exists()) {
            logger.debug("retrieve: Found file: " + file);
            filename = filename.intern();
            synchronized(filename) {
                // occassionally the BudLightWriter and the BudLightDataCenter both try to work on the
                // same file at the same time, updating while it is being read, resulting in errors
                // or possibly corrupted data. By sync-ing on the interned String filename, we keep
                // the file only in use by one at a time
            try {
                DataInputStream dis = new DataInputStream(new BufferedInputStream(new FileInputStream(file)));
                SeedRecord sr;
                int numBeforeStart = 0;
                while (true) {
                    try {
                        sr = SeedRecord.read(dis);
                    } catch(EOFException ee) {
                        // done
                        break;
                    }
                    // should only be DataRecords
                    if (sr instanceof DataRecord) {
                        DataRecord dr = (DataRecord)sr;
                        for (BtimeRange btimeRange : ranges) {
                            if (dr.getHeader().getBtimeRange().overlaps(btimeRange)) {
                                allFromFile.add(dr);
                            } else {
                                // outside request so skip
                                numBeforeStart++;
                            }
                            
                        }
                    } else {
                        throw new RuntimeException("ControlRecord in miniseed, should not happen");
                    }
                }
            } catch(Exception e) {
                logger.error("error while working on file: " + filename, e);
                throw new RuntimeException(e);
            }
            }
        }
        return allFromFile;
    }

    public String queue_seismograms(RequestFilter[] a_filterseq) throws FissuresException {
        throw new NO_IMPLEMENT();
    }

    public LocalSeismogram[] retrieve_queue(String a_request) throws FissuresException {
        throw new NO_IMPLEMENT();
    }

    public void cancel_request(String a_request) throws FissuresException {
        throw new NO_IMPLEMENT();
    }

    public String request_status(String a_request) throws FissuresException {
        throw new NO_IMPLEMENT();
    }
    
    public static Btime toBtime(Time t) {
        return FissuresConvert.getBtime(new MicroSecondDate(t));
    }

    public static List<String> formatFiles(RequestFilter[] rf) {
        Set<String> filenameSet = new HashSet<String>();
        for (int i = 0; i < rf.length; i++) {
            filenameSet.addAll(formatFiles(rf[i]));
        }
        List<String> filenameList = new ArrayList<String>();
        filenameList.addAll(filenameSet);
        Collections.sort(filenameList);
        return filenameList;
    }

    public static List<String> formatFiles(RequestFilter rf) {
        MicroSecondDate e = new MicroSecondDate(rf.end_time.date_time);
        MicroSecondDate begin = new MicroSecondDate(rf.start_time.date_time);
        return formatFiles(rf.channel_id, begin, e);
    }

    public static List<String> formatFiles(ChannelId chan, MicroSecondDate begin, MicroSecondDate e) {
        Set<String> filenameSet = new HashSet<String>();
        begin = begin.subtract(new TimeInterval(1, UnitImpl.HOUR)); // pick up data in preceeding hour that overlaps into this one

        Calendar cal = GregorianCalendar.getInstance(TimeZone.getTimeZone("GMT"));
        cal.setTime(begin);
        cal.set(Calendar.MINUTE, 0);
        cal.set(Calendar.SECOND, 0);
        cal.set(Calendar.MILLISECOND, 0);
        while (e.after(cal.getTime())) {
            String filename = BudLightWriter.formatFile(chan.network_id.network_code,
                                                      chan.station_code,
                                                      chan.site_code,
                                                      chan.channel_code,
                                              cal.get(Calendar.YEAR),
                                              cal.get(Calendar.DAY_OF_YEAR),
                                              cal.get(Calendar.HOUR_OF_DAY));
            filenameSet.add(filename);
            cal.add(Calendar.HOUR, 1);
        }
        List<String> filenameList = new ArrayList<String>();
        filenameList.addAll(filenameSet);
        Collections.sort(filenameList);
        return filenameList;
    }
    
    private static ThreadSafeSimpleDateFormat dateFormat = new ThreadSafeSimpleDateFormat("yyyy MM dd (DDD) HH:mm:ss.SSS Z");
    
    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(BudLightDataCenter.class);

    
    public File getDataDir() {
        return dataDir;
    }
}
