package edu.sc.seis.cormorant.seismogram;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.PropertyConfigurator;

import edu.iris.Fissures.model.MicroSecondDate;
import edu.iris.Fissures.model.TimeInterval;
import edu.iris.Fissures.model.UnitImpl;
import edu.sc.seis.fissuresUtil.chooser.ClockUtil;
import edu.sc.seis.fissuresUtil.exceptionHandler.DefaultExtractor;
import edu.sc.seis.fissuresUtil.exceptionHandler.ExceptionReporter;
import edu.sc.seis.fissuresUtil.exceptionHandler.Extractor;
import edu.sc.seis.fissuresUtil.exceptionHandler.GlobalExceptionHandler;
import edu.sc.seis.seisFile.mseed.Btime;
import edu.sc.seis.seisFile.mseed.DataRecord;
import edu.sc.seis.seisFile.mseed.SeedFormatException;
import edu.sc.seis.seisFile.seedlink.SeedlinkException;
import edu.sc.seis.seisFile.seedlink.SeedlinkPacket;
import edu.sc.seis.seisFile.seedlink.SeedlinkReader;

public class BudLightSeedLink implements Runnable {

    static {
        GlobalExceptionHandler.add(new DefaultExtractor() {

            public boolean canExtract(Throwable throwable) {
                return (throwable instanceof SeedFormatException);
            }

            public String extract(Throwable throwable) {
                return super.extract(throwable)+" "+((SeedFormatException)throwable).getHeader();
            }
        });
    }

    public BudLightSeedLink(DataRecordQueue queue) throws UnknownHostException, SeedlinkException, IOException {
        this(new SeedlinkReader(), queue);
    }

    public BudLightSeedLink(SeedlinkReader reader, DataRecordQueue queue) throws UnknownHostException,
            SeedlinkException, IOException {
        this.reader = reader;
        reader.setVerbose(verbose);
        this.queue = queue;
        props.load((BudLightSeedLink.class).getClassLoader().getResourceAsStream(DEFAULT_PROPS));
        PropertyConfigurator.configure(props);
        logger.info("Logging configured");
    }

    public void run() {
        logger.info("run()");
        int i = 0;
        try {
            if (commands != null) {
                for (String cmd : commands) {
                    String tcmd = cmd.trim();
                    if (tcmd.length() != 0 && !tcmd.startsWith("#")) {
                        logger.info("Cmd: "+cmd);
                        reader.sendCmd(cmd);
                    }
                }
            } else {
                reader.sendCmd("STATION " + station + " " + network);
                reader.sendCmd("SELECT " + location + channel + ".D");
                // use one of DATA and FETCH
                // DATA for realtime
                // FETCH to start flow at a specific time
                // DMC only goes back 48 hours(?) so update date to
                // something more
                // recent
                // reader.sendCmd("FETCH 0 2010,09,30,12,00,00");
                reader.sendCmd("DATA");
            }
            reader.endHandshake(); // let 'er rip
            lastPacketArrived = ClockUtil.now(); // not a real data packet, but start the clock
            while (keepGoing) {
                try {
                    while (keepGoing && (maxRecords == -1 || i < maxRecords) && reader.isConnected()) {
                        if ( ! reader.available()) {
                            // not enough bytes in stream, sleep a little
                            sleepALittle(1);
                            if (ClockUtil.now().subtract(reconnectNoData).after(lastPacketArrived)) {
                                // more than 4 minutes since last packt, shutdown the socket and try to reconnect
                                handleException(new Exception("No data in last "+reconnectNoData+", reconnecting"));
                            }
                        }
                        SeedlinkPacket slp = reader.next();
                        lastPacketArrived = ClockUtil.now();
                        DataRecord dr = slp.getMiniSeed();
                        boolean locIdFound = false;
                        if (locIdFilter == null) {
                            locIdFound = true;
                        } else {
                            for (String locId : locIdFilter) {
                                if (dr.getHeader().getLocationIdentifier().equals(locId)) {
                                    locIdFound = true;
                                    break;
                                }
                            }
                        }
                        if (locIdFound) {
                            queue.addWork(dr);
                            if (maxRecords != -1) {
                                logger.debug("packet: " + slp.getSeqNum() + " " + dr.getHeader().toString());
                            }
                            i++;
                        }
                    }
                    logger.info("finishing, wait on writer: (" + (maxRecords == -1) + " || " + (i < maxRecords)
                                + ") && " + reader.isConnected());
                    reader.close();
                    while (queue.getSize() > 0) {
                        sleepALittle(1);
                    }
                    logger.info("Done!");
                    return;
                } catch(IOException e) {
                    handleException(e);
                } catch(SeedFormatException e) {
                    handleException(e);
                } catch(SeedlinkException e) {
                    handleException(e);
                }
            }
        } catch(Exception e) {
            GlobalExceptionHandler.handle("Exception occured while using SeedLink, quiting.", e);
        }
    }
    
    public static void sleepALittle(int millies) {
        try {
            Thread.sleep(millies);
        } catch(InterruptedException e) {
            // who cares...
        }
    }
    
    public void setLocIdFilter(String[] locIdFilter) {
        this.locIdFilter = locIdFilter;
    }
    String[] locIdFilter;
    
    private void handleException(Exception e) throws SeedlinkException {
        reader.close();
        GlobalExceptionHandler.handle("Exception in SeedLink, reconnecting.", e);
        while (keepGoing && ! reader.isConnected()) {
            try {
                reader.reconnect();
            } catch(java.io.IOException ee) {
                GlobalExceptionHandler.handle("Exception reconnecting SeedLink, try again in 30 seconds.", ee);
                reader.close();
                try {
                    Thread.sleep(30000);
                } catch (InterruptedException eee) {}
            }
        }
        lastPacketArrived = ClockUtil.now(); // not a real data packet, but start the clock
    }

    public int getMaxRecords() {
        return maxRecords;
    }

    public void setMaxRecords(int m) {
        this.maxRecords = m;
    }
    
    public void stop() {
        this.keepGoing = false;
    }

    
    public List<String> getCommands() {
        return commands;
    }

    
    public void setCommands(List<String> commands) {
        this.commands = commands;
    }

    boolean keepGoing = true;

    HashMap<String, Btime> lastPacket = new HashMap<String, Btime>();

    DataRecordQueue queue;

    int maxRecords = 10;

    String station = "*";

    String network = "*";

    String location = "  ";

    String channel = "BHZ";

    List<String> commands;

    private SeedlinkReader reader;

    private static boolean verbose = false;

    Properties props = new Properties();
    
    MicroSecondDate lastPacketArrived;
    
    TimeInterval reconnectNoData = new TimeInterval(2, UnitImpl.MINUTE);

    public static List<String> loadCommands(String cmdFile) throws IOException {
        File f = new File(cmdFile);
        if (f.exists() && f.isFile() && f.canRead()) {
            BufferedReader in = new BufferedReader(new FileReader(f));
            String line;
            List<String> cmds = new ArrayList<String>();
            while ((line = in.readLine()) != null) {
                cmds.add(line);
            }
            return cmds;
        } else {
            logger.error("Unable to read from " + cmdFile);
            throw new IOException("Unable to read from " + cmdFile);
        }
    }
    
    public static void main(String[] args) throws Exception {
        String dataDir = "Data";
        DataRecordQueue queue = new DataRecordQueue();
        BudLightSeedLink bud = new BudLightSeedLink(queue);
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-h") || args[i].equals("--help")) {
                System.out.println("Usage:");
                System.out.println("-h         help");
                System.out.println("-n net     network");
                System.out.println("-s sta     station");
                System.out.println("-l loc     loc id");
                System.out.println("-c chan    channel");
                System.out.println("-m int     max records before quiting (-1 to run forever)");
                System.out.println("-f cmdfile        seedlink commands file to use instead of nslc from above");
                System.out.println("-d dir     data directory for budlight");
            }
            if (i < args.length - 1) {
                if (args[i].equals("-n")) {
                    bud.network = args[i + 1];
                    i++;
                }
                if (args[i].equals("-s")) {
                    bud.station = args[i + 1];
                    i++;
                }
                if (args[i].equals("-l")) {
                    bud.location = args[i + 1];
                    i++;
                }
                if (args[i].equals("-c")) {
                    bud.channel = args[i + 1];
                    i++;
                }
                if (args[i].equals("-m")) {
                    bud.setMaxRecords(Integer.parseInt(args[i + 1]));
                    i++;
                }
                if (args[i].equals("-f")) {
                    bud.commands = loadCommands(args[i + 1]);
                }
                if (args[i].equals("-d")) {
                    dataDir = args[i + 1];
                    i++;
                }
            }
        }
        BudLightWriter bdw = new BudLightWriter(new File(dataDir), queue);
        Thread t = new Thread(bdw);
        t.setDaemon(true);
        t.start();
        bud.run();
    }

    private static final org.slf4j.Logger logger = org.slf4j.LoggerFactory.getLogger(BudLightSeedLink.class);

    static final String DEFAULT_PROPS = "edu/sc/seis/cormorant/seismogram/budlight.prop";
}
